@@ -342,6 +342,8 @@ def params_to_fa(params, mask=None):
         mask = np.ones(params.shape[0:-1]).astype(bool)
     evals, _ = np.linalg.eigh(np.nan_to_num(params_to_D(params[mask])))
     avg_evals = np.mean(evals, axis=-1)
+    sum_sq_evals = np.sum(evals ** 2, axis=-1)
+    sum_sq_evals[sum_sq_evals == 0] = np.nan  # To avoid warnings for dividing by zero
     fa = np.zeros(mask.shape)
     fa[mask] = np.sqrt(
         1.5
@@ -350,7 +352,7 @@ def params_to_fa(params, mask=None):
             + (evals[..., 1] - avg_evals) ** 2
             + (evals[..., 2] - avg_evals) ** 2
         )
-        / np.sum(evals ** 2, axis=-1)
+        / sum_sq_evals
     )
     return fa
 
@@ -680,7 +682,7 @@ def _nlls_fit(data, design_matrix, mask=None):
 
 @jax.jit
 def _akc_mask(W, vs, mask):
-    """Compute a mask based on condition AKC > 0 along all directions in `vs`.
+    """Compute a mask based on condition AKC >= 0 along all directions in `vs`.
 
     Parameters
     ----------
@@ -697,7 +699,7 @@ def _akc_mask(W, vs, mask):
     """
     akc_mask = np.ones(mask.shape)
     for v in vs:
-        akc_mask *= (v.T @ (v.T @ W @ v) @ v) > 0
+        akc_mask *= (v.T @ (v.T @ W @ v) @ v) >= 0
     akc_mask *= mask
     return akc_mask
 
@@ -727,7 +729,7 @@ def _predict(data, m, akc_mask, seed, mask=None):
     if mask is None:
         mask = np.ones(data.shape[0:-1]).astype(bool)
     X = data[akc_mask]
-    y = np.clip(np.nan_to_num(m[akc_mask]), MIN_K, MAX_K)
+    y = m[akc_mask]
     reg = MLPRegressor(
         hidden_layer_sizes=(20, 20), max_iter=int(1e3), random_state=seed,
     ).fit(X, y)
@@ -755,7 +757,9 @@ def signal(params, design_matrix, mask=None):
     """
     if mask is None:
         mask = np.ones(params.shape[0:-1]).astype(bool)
-    S_hat = np.exp(design_matrix @ params[..., np.newaxis])[..., 0]
+    XB = design_matrix @ params[..., np.newaxis]
+    XB[XB > 200] = np.inf  # To avoid overflow warnings
+    S_hat = np.exp(XB)[..., 0]
     S_hat[~mask] = 0
     return S_hat
 
@@ -797,7 +801,6 @@ def _calculate_x0(S0, D, mtk_pred, ak_pred, rk_pred, mask=None):
 
     @numba.njit
     def p(u):
-        """Return tensor P."""
         P = np.zeros((3, 3, 3, 3))
         for i in range(3):
             for j in range(3):
@@ -808,7 +811,6 @@ def _calculate_x0(S0, D, mtk_pred, ak_pred, rk_pred, mask=None):
 
     @numba.njit
     def d(i, j):
-        """Kronecker delta."""
         if i == j:
             return 1
         else:
@@ -816,7 +818,6 @@ def _calculate_x0(S0, D, mtk_pred, ak_pred, rk_pred, mask=None):
 
     @numba.njit
     def q(u):
-        """Return tensor Q."""
         Q = np.zeros((3, 3, 3, 3))
         for i in range(3):
             for j in range(3):
@@ -843,27 +844,21 @@ def _calculate_x0(S0, D, mtk_pred, ak_pred, rk_pred, mask=None):
 
     x0_flat = np.zeros((size, 22))
     for i in range(size):
-
         u = evecs[i, :, 2]
-
         md = np.mean(evals[i, :])
         ad = evals[i, 2]
         rd = np.mean(evals[i, 0:2])
         if rd < 0.1:
             rd = 0.1
-
         mtk = mtk_pred_flat[i]
         atk = ak_pred_flat[i] * ad ** 2 / md ** 2
         rtk = rk_pred_flat[i] * rd ** 2 / md ** 2
-
         D = rd * np.eye(3) + (ad - rd) * u[:, np.newaxis] @ u[np.newaxis, :]
-
         W = (
             0.5 * (10 * rtk + 5 * atk - 15 * mtk) * p(u)
             + rtk * I
             + 1.5 * (5 * mtk - atk - 4 * rtk) * q(u)
         )
-
         x0_flat[i, 0] = np.log(S0_flat[i])
         x0_flat[i, 1] = D[0, 0]
         x0_flat[i, 2] = D[1, 1]
@@ -927,22 +922,31 @@ def _reg_nlls_fit(
         Floating-point array with shape (..., 1, 3).
     radial_dirs : numpy.ndarray
         Floating-point array with shape (..., 10, 3).
-    alpha : float
+    alpha : float, optional
         Constant controlling regularization term magnitude.
     mask : numpy.ndarray, optional
         Boolean array.
-    quiet : bool
+    quiet : bool, optional
         Whether not to print messages about computation progress.
 
     Returns
     -------
     params : numpy.ndarray
-    fit_status : numpy.ndarray
+    status : numpy.ndarray
     """
 
     if mask is None:
         mask = np.ones(data.shape[0:-1]).astype(bool)
 
+    if alpha is None:
+        mse_dki = np.median(
+            np.mean((signal(x0[mask], design_matrix) - data[mask]) ** 2, axis=1)
+        )
+        mse_mk = np.median((mk_pred[mask] - params_to_mk(x0[mask])) ** 2)
+        alpha = 0.1 * mse_dki / mse_mk
+    if not quiet:
+        print(f"alpha = {np.round(alpha, 5)}")
+
     data_flat = jnp.asarray(data[mask])
     design_matrix = jnp.asarray(design_matrix)
     x0_flat = jnp.asarray(x0[mask])
@@ -954,17 +958,6 @@ def _reg_nlls_fit(
     radial_dirs_flat = jnp.asarray(radial_dirs[mask])
     size = len(data_flat)
 
-    if alpha is None:
-        mse_dki = np.median(
-            np.mean((signal(x0_flat, design_matrix) - data_flat) ** 2, axis=1)
-        )
-        mse_mk = np.median((mk_pred[mask] - params_to_mk(x0_flat)) ** 2)
-        alpha = 0.1 * mse_dki / mse_mk
-    if not quiet:
-        print(f"alpha = {alpha}")
-
-    rd_flat = jnp.asarray(params_to_rd(x0, mask)[mask])
-
     @jax.jit
     def cost(
         params,
@@ -977,7 +970,6 @@ def _reg_nlls_fit(
         rk_pred,
         axial_dir,
         radial_dirs,
-        rd,
     ):
         return (
             jnp.mean((jnp.exp(design_matrix @ params) - y) ** 2)
@@ -986,9 +978,6 @@ def _reg_nlls_fit(
             + alpha * (jnp.mean(_akc(params, radial_dirs)) - rk_pred) ** 2
         )
 
-    jac = jax.jit(jax.jacfwd(cost))
-    hess = jax.jit(jax.jacfwd(jax.jacrev(cost)))
-
     @jax.jit
     def jit_minimize(i):
         return minimize(
@@ -1004,73 +993,68 @@ def _reg_nlls_fit(
                 rk_pred_flat[i],
                 axial_dirs_flat[i],
                 radial_dirs_flat[i],
-                rd_flat[i],
             ),
             method="BFGS",
             options={
                 "maxiter": int(1e4),
-                "line_search_maxiter": int(1e3),
-                "gtol": 1e-3,
+                "line_search_maxiter": int(1e4),
+                "gtol": 1e-4,
             },
         )
 
-    fit_status_flat = np.zeros(size)
     params_flat = np.zeros((size, 22))
+    status_flat = np.zeros(size)
     for i in range(size):
         if not quiet:
             print(f"{int(i/size*100)}%", end="\r")
         results = jit_minimize(i)
         params_flat[i] = results.x
-        fit_status_flat[i] = results.status
+        status_flat[i] = results.status
+    if not quiet:
+        print("100%")
 
     params = np.zeros(mask.shape + (22,))
     params[mask] = params_flat
-    fit_status = np.zeros(mask.shape)
-    fit_status[mask] = fit_status_flat
+    status = np.zeros(mask.shape)
+    status[mask] = status_flat
 
-    return params, fit_status
+    return params, status
 
 
-def fit(data, bvals, bvecs, mask=None, alpha=None, seed=123, quiet=False):
-    """Estimate diffusion and kurtosis tensor elements.
+class FitResult:
+    """Class for storing the regularized NLLS fit results.
 
-    This function does the following:
-
-        1. Remove infinities, nans, and negative values, and scale data and
-           b-values.
-        2. Estimate DKI model parameters using standard NLLS.
-        3. Train a multilayer perceptron to predict kurtosis measures from data
-           in voxels where the apparent kurtosis coefficient estimated by the
-           NLLS fit was > 0 and < 10 along all directions.
-        4. Estimate model parameters using regularized NLLS where the
-           regularization terms increase the objective function value when
-           MK, AK, and RK deviate from their predicted values. Axially
-           symmetric tensors with plausible magnitudes are used as initial
-           positions for the fit.
-
-    Parameters
+    Attributes
     ----------
-    data : numpy.ndarray
-        Floating-point array with shape (..., number of acquisitions).
-    bvals : numpy.ndarray
-        Floating-point array with shape (number of acquisitions,).
-    bvecs : numpy.ndarray
-        Floating-point array with shape (number of acquisitions, 3).
-    mask : numpy.ndarray, optional
-        Boolean array with the same shape as `data` without the last dimension.
-    alpha : float, optional
-        Constant controlling regularization term magnitudes in the objective
-        function.
-    seed : int
-        Seed for pseudo-random number generation to initialize neural network
-        weights.
-    quiet : bool
-        Whether not to print messages about computation progress.
-
-    Returns
-    -------
     params : numpy.ndarray
-        Estimated model parameters:
+        Estimated model parameters.
+    status : numpy.ndarray
+        Codes returned by the optimizer in each voxel where it was run. 0 means
+        converged succesfully, 1 means maximum BFGS iterations reached, 3 means
+        zoom failed, 4 means saddle point reached, 5 means maximum line search
+        iterations reached, and -1 means undefined.
+    mk_pred : numpy.ndarray
+        Predicted mean kurtosis map.
+    ak_pred : numpy.ndarray
+        Predicted axial kurtosis map.
+    rk_pred : numpy.ndarray
+        Predicted radial kurtosis map.
+    mask : numpy.ndarray
+        Mask defining voxels in which the fit was run.
+    params_nlls : numpy.ndarray
+        Parameter estimates by the standard NLLS fit.
+    akc_mask : nump.ndarray
+        Mask defining voxels in which AKC computed from the standard NLLS fit
+        results was non-negative along all directions.
+    x0 : numpy.ndarray
+        Initial positions of the regularized NLLS fit.
+    alpha : float
+        Value of the constant controlling the magnitude of regularization terms
+        in the regularized NLLS fit.
+
+    Notes
+    -----
+    The model parameters are the following:
 
             params[..., 0] = log(S0)
             params[..., 1] = D_xx
@@ -1094,61 +1078,111 @@ def fit(data, bvals, bvecs, mask=None, alpha=None, seed=123, quiet=False):
             params[..., 19] = W_xxyz * MD ** 2
             params[..., 20] = W_yyxz * MD ** 2
             params[..., 21] = W_zzxy * MD ** 2
-       
-    fit_status : numpy.ndarray
-        Codes returned by the optimizer in each voxel where it was run. 0 means
-        converged succesfully, 1 means maximum BFGS iterations reached, 3 means
-        zoom failed, 4 means saddle point reached, 5 means maximum line search
-        iterations reached, and -1 means undefined.
-    mk_pred : numpy.ndarray
-        The predicted mean kurtosis map.
-    ak_pred : numpy.ndarray
-        The predicted axial kurtosis map.
-    rk_pred : numpy.ndarray
-        The predicted radial kurtosis map.
     """
+
+    def __init__(
+        self,
+        params,
+        status,
+        mk_pred,
+        ak_pred,
+        rk_pred,
+        mask,
+        params_nlls,
+        akc_mask,
+        x0,
+        alpha,
+    ):
+        self.params = params
+        self.status = status
+        self.mk_pred = mk_pred
+        self.ak_pred = ak_pred
+        self.rk_pred = rk_pred
+        self.mask = mask
+        self.params_nlls = params_nlls
+        self.akc_mask = akc_mask
+        self.x0 = x0
+        self.alpha = alpha
+        return
+
+
+def fit(data, bvals, bvecs, mask=None, alpha=None, seed=123, quiet=False):
+    """Estimate diffusion and kurtosis tensor elements from data.
+
+    This function does the following:
+
+        1. Remove infinities, nans, and negative values in `data`, and scale
+           values in `data` and `bvals`.
+        2. Estimate model parameters using standard NLLS.
+        3. Train multilayer perceptrons to predict kurtosis measures from data
+           in voxels where the apparent kurtosis coefficient computed from the
+           NLLS fit results is non-negative along all directions.
+        4. Estimate model parameters using regularized NLLS where the
+           regularization terms increase the objective function value when
+           MK, AK, and RK deviate from their predicted values. Axially
+           symmetric tensors with plausible magnitudes are used as initial
+           positions for the fit.
+
+    Parameters
+    ----------
+    data : numpy.ndarray
+        Floating-point array with shape (..., number of acquisitions).
+    bvals : numpy.ndarray
+        Floating-point array with shape (number of acquisitions,).
+    bvecs : numpy.ndarray
+        Floating-point array with shape (number of acquisitions, 3).
+    mask : numpy.ndarray, optional
+        Boolean array with the same shape as `data` without the last dimension.
+    alpha : float, optional
+        Constant controlling regularization term magnitudes in the objective
+        function. If not given, `alpha` is equal to the 0.1 times median
+        squared error of the standard NLLS fit divided by the median squared
+        error of the MK prediction.
+    seed : int, optional
+        Seed for pseudo-random number generation to initialize neural network
+        weights.
+    quiet : bool, optional
+        Whether not to print messages about computation progress.
+
+    Returns
+    -------
+    dkmri.FitResult
+    """
+
     if mask is None:
         mask = np.ones(data.shape[0:-1]).astype(bool)
 
+    data = data.astype(float)
+    C_data = np.nanmean(data[mask][..., np.where(bvals == np.min(bvals))])
+    data = data / C_data
     data[np.isinf(data)] = np.nan
     data[np.isnan(data)] = 0
-    C_data = np.mean(data[..., np.where(bvals == np.min(bvals))])
-    data /= C_data
     min_signal = np.finfo(float).resolution
     data[data < min_signal] = min_signal
+
     C_bvals = np.mean(bvals)
-    bvals /= C_bvals
+    bvals = bvals / C_bvals
 
     if not quiet:
         print("Fitting DKI to data with standard NLLS")
     X = design_matrix(bvals, bvecs)
-    nlls_params = _nlls_fit(data, X, mask)
+    params_nlls = _nlls_fit(data, X, mask)
 
-    akc_mask = _akc_mask(params_to_W(nlls_params), _45_dirs, mask).astype(bool)
     if not quiet:
-        print("Training a MLP to predict MK")
-    mk = np.clip(params_to_mk(nlls_params, mask), MIN_K, MAX_K)
+        print("Training neural networks to predict kurtosis maps")
+    akc_mask = _akc_mask(params_to_W(params_nlls), _45_dirs, mask).astype(bool)
+    mk = np.clip(np.nan_to_num(params_to_mk(params_nlls, mask)), MIN_K, MAX_K)
     mk_pred, R2 = _predict(data, mk, akc_mask, seed, mask)
     if not quiet:
-        print(f"R^2 on training data for MK = {R2}")
-    if not quiet:
-        print("Training a MLP to predict MTK")
-    mtk = np.clip(_mtk(nlls_params, mask), MIN_K, MAX_K)
-    mtk_pred, R2 = _predict(data, mtk, akc_mask, seed, mask)
-    if not quiet:
-        print(f"R^2 on training data for MTK = {R2}")
-    if not quiet:
-        print("Training a MLP to predict AK")
-    ak = np.clip(params_to_ak(nlls_params, mask), MIN_K, MAX_K)
+        print(f"R^2 = {np.round(R2, 5)} for MK")
+    ak = np.clip(np.nan_to_num(params_to_ak(params_nlls, mask)), MIN_K, MAX_K)
     ak_pred, R2 = _predict(data, ak, akc_mask, seed, mask)
     if not quiet:
-        print(f"R^2 on training data for AK = {R2}")
-    if not quiet:
-        print("Training a MLP to predict RK")
-    rk = np.clip(params_to_rk(nlls_params, mask), MIN_K, MAX_K)
+        print(f"R^2 = {np.round(R2, 5)} for AK")
+    rk = np.clip(np.nan_to_num(params_to_rk(params_nlls, mask)), MIN_K, MAX_K)
     rk_pred, R2 = _predict(data, rk, akc_mask, seed, mask)
     if not quiet:
-        print(f"R^2 on training data for RK = {R2}")
+        print(f"R^2 = {np.round(R2, 5)} for RK")
 
     if not quiet:
         print("Calculating initial positions")
@@ -1157,7 +1191,7 @@ def fit(data, bvals, bvecs, mask=None, alpha=None, seed=123, quiet=False):
     radial_dirs = np.zeros(mask.shape + (10, 3))
     radial_dirs_flat = radial_dirs[mask]
     size = len(radial_dirs_flat)
-    _, evecs = np.linalg.eigh(np.nan_to_num(params_to_D(nlls_params)))
+    _, evecs = np.linalg.eigh(np.nan_to_num(params_to_D(params_nlls)))
     evecs_flat = evecs[mask]
     for i in range(size):
         axial_dirs_flat[i] = evecs_flat[i, :, 2].T
@@ -1165,13 +1199,15 @@ def fit(data, bvals, bvecs, mask=None, alpha=None, seed=123, quiet=False):
         radial_dirs_flat[i] = (R @ _10_dirs.T).T
     axial_dirs[mask] = axial_dirs_flat
     radial_dirs[mask] = radial_dirs_flat
-    S0 = np.exp(nlls_params[..., 0])
-    D = params_to_D(nlls_params)
+    S0 = np.exp(params_nlls[..., 0])
+    D = params_to_D(params_nlls)
+    mtk = np.clip(np.nan_to_num(_mtk(params_nlls, mask)), MIN_K, MAX_K)
+    mtk_pred, _ = _predict(data, mtk, akc_mask, seed, mask)
     x0 = _calculate_x0(S0, D, mtk_pred, ak_pred, rk_pred, mask)
 
     if not quiet:
         print("Fitting DKI to data with regularized NLLS")
-    params, fit_status = _reg_nlls_fit(
+    params, status = _reg_nlls_fit(
         data,
         X,
         x0,
@@ -1189,16 +1225,38 @@ def fit(data, bvals, bvecs, mask=None, alpha=None, seed=123, quiet=False):
     params[..., 1:7] /= C_bvals
     params[..., 7::] /= C_bvals ** 2
 
-    return params, fit_status, mk_pred, ak_pred, rk_pred
+    params_nlls[..., 0] += np.log(C_data)
+    params_nlls[..., 1:7] /= C_bvals
+    params_nlls[..., 7::] /= C_bvals ** 2
+
+    x0[..., 0] += np.log(C_data)
+    x0[..., 1:7] /= C_bvals
+    x0[..., 7::] /= C_bvals ** 2
+
+    return FitResult(
+        params,
+        status,
+        mk_pred,
+        ak_pred,
+        rk_pred,
+        mask,
+        params_nlls,
+        akc_mask,
+        x0,
+        alpha,
+    )
 
 
 if __name__ == "__main__":
 
-    # Parse arguments
-
     parser = argparse.ArgumentParser(
         description=(
-            "Estimate diffusion and kurtosis tensors, and compute parameter maps"
+            """Estimate diffusion kurtosis imaging parameters and compute parameter maps.
+            The command for using dkmri.py is "dkmri.py data bvals bvecs
+            optional-arguments", where data, bvals, and bvecs are the paths of the files
+            containing the diffusion-weighted data, b-values, and b-vectors, and
+            optional-arguments is where to define things such as which parameter maps to
+            save. Visit https://github.com/kerkelae/dkmri for more info."""
         )
     )
     parser.add_argument(
@@ -1213,8 +1271,7 @@ if __name__ == "__main__":
     parser.add_argument(
         "-mask",
         help=(
-            "path of a NIfTI file with a binary mask definining the voxels where to "
-            "estimate the tensors"
+            "path of a NIfTI file with a mask definining where to estimate parameters"
         ),
     )
     parser.add_argument(
@@ -1255,7 +1312,8 @@ if __name__ == "__main__":
         help="path of a NIfTI file in which to save the predicted radial kurtosis map",
     )
     parser.add_argument(
-        "-status", help="path of a NIfTI file in which to save the fit status map",
+        "-status",
+        help="path of a NIfTI file in which to save the codes returned by the optimizer",
     )
     parser.add_argument(
         "-params",
@@ -1263,12 +1321,10 @@ if __name__ == "__main__":
     )
     args = parser.parse_args()
 
-    # Load data
-
     data_img = nib.load(args.data)
     data = data_img.get_fdata()
     affine = data_img.affine
-    bvals = np.loadtxt(args.bvals) * 1e-3
+    bvals = np.loadtxt(args.bvals)
     bvecs = np.loadtxt(args.bvecs)
     if bvecs.ndim == 2 and bvecs.shape[0] == 3:
         bvecs = bvecs.T
@@ -1277,36 +1333,45 @@ if __name__ == "__main__":
     else:
         mask = None
 
-    # Fit model to data with regularized NLLS
+    fit_result = fit(data, bvals, bvecs, mask)
 
-    params, fit_status, mk_pred, ak_pred, rk_pred = fit(data, bvals, bvecs, mask)
-
-    # Save results
-
-    if args.params:
-        nib.save(nib.Nifti1Image(params, affine), args.params)
     if args.md:
-        nib.save(nib.Nifti1Image(params_to_md(params, mask) * 1e-3, affine), args.md)
+        nib.save(
+            nib.Nifti1Image(params_to_md(fit_result.params, mask), affine), args.md,
+        )
     if args.ad:
-        nib.save(nib.Nifti1Image(params_to_ad(params, mask) * 1e-3, affine), args.ad)
+        nib.save(
+            nib.Nifti1Image(params_to_ad(fit_result.params, mask), affine), args.ad,
+        )
     if args.rd:
-        nib.save(nib.Nifti1Image(params_to_rd(params, mask) * 1e-3, affine), args.rd)
+        nib.save(
+            nib.Nifti1Image(params_to_rd(fit_result.params, mask), affine), args.rd,
+        )
     if args.fa:
-        nib.save(nib.Nifti1Image(params_to_fa(params, mask) * 1e-3, affine), args.fa)
+        nib.save(
+            nib.Nifti1Image(params_to_fa(fit_result.params, mask), affine), args.fa,
+        )
     if args.mk:
-        nib.save(nib.Nifti1Image(params_to_mk(params, mask), affine), args.mk)
+        nib.save(
+            nib.Nifti1Image(params_to_mk(fit_result.params, mask), affine), args.mk
+        )
     if args.ak:
-        nib.save(nib.Nifti1Image(params_to_ak(params, mask), affine), args.ak)
+        nib.save(
+            nib.Nifti1Image(params_to_ak(fit_result.params, mask), affine), args.ak
+        )
     if args.rk:
-        nib.save(nib.Nifti1Image(params_to_rk(params, mask), affine), args.rk)
+        nib.save(
+            nib.Nifti1Image(params_to_rk(fit_result.params, mask), affine), args.rk
+        )
     if args.s0:
-        nib.save(nib.Nifti1Image(np.exp(params[..., 0]), affine), args.s0)
+        nib.save(nib.Nifti1Image(np.exp(fit_result.params[..., 0]), affine), args.s0)
     if args.mk_pred:
-        nib.save(nib.Nifti1Image(mk_pred, affine), args.mk_pred)
+        nib.save(nib.Nifti1Image(fit_result.mk_pred, affine), args.mk_pred)
     if args.ak_pred:
-        nib.save(nib.Nifti1Image(ak_pred, affine), args.ak_pred)
+        nib.save(nib.Nifti1Image(fit_result.ak_pred, affine), args.ak_pred)
     if args.rk_pred:
-        nib.save(nib.Nifti1Image(rk_pred, affine), args.rk_pred)
-    if args.fit_status:
-        nib.save(nib.Nifti1Image(fit_status, affine), args.fit_status)
-
+        nib.save(nib.Nifti1Image(fit_result.rk_pred, affine), args.rk_pred)
+    if args.status:
+        nib.save(nib.Nifti1Image(fit_result.status, affine), args.status)
+    if args.params:
+        nib.save(nib.Nifti1Image(fit_result.params, affine), args.params)
